# -*- coding: utf-8 -*-
# synadm
# Copyright (C) 2023 Nicolas Peugnet
#
# synadm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# synadm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Alternative entry point for running synadm

synadm is typically installed via pip, utilizing setuptools-configured entry
points for accessibility as `synadm`.

This file allows execution without formal installation using
`python -m synadm`, which proves useful, for instance, in Debian GNU/Linux
packaging.
"""

from synadm.cli import root

root()



# -*- coding: utf-8 -*-
# synadm
# Copyright (C) 2023-2025 Johannes Tiefenbacher
#
# synadm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# synadm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"Common CLI options, option groups, helpers and utilities."

import click

from click_option_group import MutuallyExclusiveOptionGroup


def common_opts_raw_command(function):
    return click.argument(
        "endpoint", type=str
    )(
        click.option(
            "--method", "-m",
            type=click.Choice(["get", "post", "put", "delete"]),
            help="The HTTP method used for the request.",
            default="get", show_default=True
        )(function)
    )


data_group_raw_command = MutuallyExclusiveOptionGroup(
    "Data",
    help=""
)


def data_opts_raw_command(function):
    return data_group_raw_command.option(
            "--data", "-d", type=str, default='{}', show_default=True,
            help="""The JSON string sent in the body of post, put and delete
            requests - provided as a string. Make sure to escape it from shell
            interpretation by using single quotes. E.g '{"key1": "value1",
            "key2": 123}'"""
    )(
        data_group_raw_command.option(
            "--data-file", "-f", type=click.File("rt"),
            show_default=True,
            help="""Read JSON data from file. To read from stdin use "-" as the
            filename argument."""
        )(function)
    )



# -*- coding: utf-8 -*-
# synadm
# Copyright (C) 2020-2025 Johannes Tiefenbacher
# Copyright (C) 2023-2025 JacksonChen666
#
# synadm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# synadm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Room-related CLI commands
"""

import click
from click_option_group import MutuallyExclusiveOptionGroup, optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup

from synadm import cli


@cli.root.group()
def room():
    """ Manipulate rooms and room membership.

    The syntax of room IDs in synadm is as the Matrix spec defines:
    https://spec.matrix.org/latest/#room-structure.
    Make sure to escape the ! character from your shell. In bash and zsh this
    can be achieved by using single quotes ('), eg. '!id123abc:matrix.DOMAIN'
    """


@room.command()
@click.argument("room_id_or_alias", type=str)
@click.argument("user_id", type=str)
@click.pass_obj
def join(helper, room_id_or_alias, user_id):
    """ Join a room.

    The administrator must belong to the room and have sufficient privilege to
    invite the user. If this is not the case, you can use the make-admin
    command.
    """
    mxid = helper.generate_mxid(user_id)
    out = helper.api.room_join(room_id_or_alias, mxid)
    helper.output(out)


@room.command()
@click.argument("room_id_or_alias", type=str)
@click.option(
    "--reverse", "-r", is_flag=True, default=False, show_default=True,
    help="""Fetch all room aliases corresponding to a given room ID,
    instead of the other way round.
    """)
@click.pass_obj
def resolve(helper, room_id_or_alias, reverse):
    """ Lookup room ID from alias or vice versa.
    """
    if reverse:
        out = helper.matrix_api.room_get_aliases(room_id_or_alias)
    else:
        out = helper.matrix_api.room_get_id(room_id_or_alias)

    if out is None:
        click.echo("Room resolve failed.")
        raise SystemExit(1)
    helper.output(out)


@room.command(name="list")
@click.pass_obj
@click.option(
    "--from", "-f", "from_", type=int, default=0, show_default=True,
    help="""Offset room listing by given number. This option is used for
    pagination.""")
@click.option(
    "--limit", "-l", type=int, default=100, show_default=True,
    help="Maximum amount of rooms to return.")
@click.option(
    "--name", "-n", type=str,
    help="""Filter rooms by parts of their room name, canonical alias and room
    id.""")
@click.option(
    "--sort", "-s", type=click.Choice(
        ["name", "canonical_alias", "joined_members", "joined_local_members",
         "version", "creator", "encryption", "federatable", "public",
         "join_rules", "guest_access", "history_visibility", "state_events"]),
    help="The method in which to sort the returned list of rooms.")
@click.option(
    "--reverse", "-r", is_flag=True, default=False,
    help="""Direction of room order. If set it will reverse the sort order of
    --order-by method.""")
@click.option(
    "--empty/--not-empty", "-e/-E", "empty_rooms", is_flag=True,
    type=bool, default=None, help="""Query for empty rooms only, or not
    empty rooms only. If unspecified, both empty and non-empty rooms are
    included.""")
def list_room_cmd(helper, from_, limit, name, sort, reverse, empty_rooms):
    """ List and search for rooms.
    """
    rooms = helper.api.room_list(from_, limit, name, sort, reverse,
                                 empty_rooms)
    if rooms is None:
        click.echo("Rooms could not be fetched.")
        raise SystemExit(1)
    if helper.output_format == "human":
        if int(rooms["total_rooms"]) != 0:
            helper.output(rooms["rooms"])
        if "next_batch" in rooms:
            click.echo("There are more rooms than shown, use '--from {}'"
                       .format(rooms["next_batch"]))
    else:
        helper.output(rooms)


@room.command()
@click.argument("room_id", type=str)
@click.pass_obj
def details(helper, room_id):
    """ Get room details.
    """
    room_details = helper.api.room_details(room_id)
    if room_details is None:
        click.echo("Room details could not be fetched.")
        raise SystemExit(1)
    helper.output(room_details)


@room.command()
@click.argument("room_id", type=str)
@click.pass_obj
def state(helper, room_id):
    """ Get a list of all state events in a room.
    """
    room_state = helper.api.room_state(room_id)
    if room_state is None:
        click.echo("Room state could not be fetched.")
        raise SystemExit(1)
    helper.output(room_state)


@room.command()
@click.option(
    "--room-id", "-i", type=str,
    help="""View power levels of this room only.""")
@click.option(
    "--all-details", "-a", is_flag=True, default=False,
    help="""Show detailed information about each room. The default is to only
    show room_id, name, canonical_alias and power_levels.""")
@click.option(
    "--from", "-f", "from_", type=int, default=0, show_default=True,
    help="""Offset room listing by given number. This option is used for
    pagination.""")
@click.option(
    "--limit", "-l", type=int, default=10, show_default=True,
    help="Maximum amount of rooms to return.")
@click.option(
    "--name", "-n", type=str,
    help="""Filter rooms by parts of their room name, canonical alias and room
    id.""")
@click.option(
    "--sort", "-s", type=click.Choice(
        ["name", "canonical_alias", "joined_members", "joined_local_members",
         "version", "creator", "encryption", "federatable", "public",
         "join_rules", "guest_access", "history_visibility", "state_events"]),
    help="The method in which to sort the returned list of rooms.")
@click.option(
    "--reverse", "-r", is_flag=True, default=False,
    help="""Direction of room order. If set it will reverse the sort order of
    --order-by method.""")
@click.pass_obj
def power_levels(helper, room_id, all_details, from_, limit, name, sort,
                 reverse):
    """ List user's power levels set in rooms.

    A combination of commands `room list` and `room state`. It enriches
    the room list response with a list of users and their corresponding
    power levels set. It only displays a subset of the available information
    (room name, id, aliases and power levels). Increase the number of rooms
    fetched using --limit/-l (default: 10) or use the pagination option
    --from/-f to go beyond the default. Use --name/-n to search. This command
    can require quite some time to complete depending on those options.
    """
    rooms_power = helper.api.room_power_levels(
        from_, limit, name, sort, reverse, room_id, all_details,
        helper.output_format)
    if rooms_power is None:
        click.echo("Rooms and power levels could not be fetched.")
        raise SystemExit(1)

    if helper.output_format == "human":
        if int(rooms_power["total_rooms"]) != 0:
            helper.output(rooms_power["rooms"])
            click.echo("Rooms with power levels found in current batch: {}"
                       .format(rooms_power["rooms_w_power_levels_curr_batch"]))
            click.echo("Total rooms: {}"
                       .format(rooms_power["total_rooms"]))
        if "next_batch" in rooms_power:
            click.echo("Use '--from/-f {}' to view next batch."
                       .format(rooms_power["next_batch"]))
    else:
        helper.output(rooms_power)


@room.command()
@click.argument("room_id", type=str)
@click.pass_obj
def members(helper, room_id):
    """ List current room members.
    """
    room_members = helper.api.room_members(room_id)
    if room_members is None:
        click.echo("Room members could not be fetched.")
        raise SystemExit(1)
    if helper.output_format == "human":
        click.echo("Total members in room: {}"
                   .format(room_members["total"]))
        if int(room_members["total"]) != 0:
            helper.output(room_members["members"])
    else:
        helper.output(room_members)


@room.command()
@click.argument("room_id", type=str)
@click.option(
    "--new-room-user-id", "-u", type=str,
    help="""If set, a new room will be created with this user ID as the
    creator and admin, and all users in the old room will be moved into
    that room. If not set, no new room will be created and the users will
    just be removed from the old room. The user ID must be on the local
    server, but does not necessarily have to belong to a registered
    user.""")
@click.option(
    "--room-name", "-n", type=str,
    help="""A string representing the name of the room that new users will
    be invited to. Defaults to "Content Violation Notification" """)
@click.option(
    "--message", "-m", type=str,
    help="""A string containing the first message that will be sent as
    new_room_user_id in the new room. Ideally this will clearly convey why
    the original room was shut down. Defaults to "Sharing illegal content
    on this server is not permitted and rooms in violation will be
    blocked." """)
@click.option(
    "--block", "-b", is_flag=True, default=False, show_default=True,
    help="""If set, this room will be added to a blocking list,
    preventing future attempts to join the room""")
@click.option(
    "--no-purge", is_flag=True, default=False, show_default=True,
    help="""Prevent removing of all traces of the room from your
    database.""")
@click.option(
    "--force-purge", is_flag=True, default=False, show_default=True,
    help="""Force a purge to go ahead even if there are local users still
    in the room. Do not use this unless a regular purge operation fails,
    as it could leave those users' clients in a confused state.""")
@click.option(
    "--v1", is_flag=True, default=False, show_default=True,
    help="""Use version 1 of the room delete API instead of version 2, which
    will wait until the room deletion is complete.""")
@click.pass_obj
@click.pass_context
def delete(ctx, helper, room_id, new_room_user_id, room_name, message, block,
           no_purge, force_purge, v1):
    """ Delete and possibly purge a room.

    By default, the v2 API is used, which will return a delete_id and delete
    the room asynchronously instead of waiting for a room to be deleted (the
    v1 API will wait for the room to be deleted). You can check the status
    of the room deletion with synadm room delete-status.
    """
    if no_purge and force_purge:
        click.echo("--force-purge will be ignored as --no-purge is set")
    room_details = helper.api.room_details(room_id)
    if "errcode" in room_details.keys():
        if room_details["errcode"] == "M_NOT_FOUND":
            click.echo("Room not found.")
            raise SystemExit(1)
        else:
            click.echo("Unrecognized error")
            helper.output(room_details)
            raise SystemExit(1)
    helper.output(room_details)
    ctx.invoke(members, room_id=room_id)
    sure = (
        helper.no_confirm or
        click.prompt("Are you sure you want to delete this room? (y/N)",
                     type=bool, default=False, show_default=False)
    )
    if sure:
        mxid = helper.generate_mxid(new_room_user_id)
        if v1:
            room_del = helper.api.room_delete(
                room_id, mxid, room_name,
                message, block, no_purge, force_purge)
        else:
            room_del = helper.api.room_delete_v2(
                room_id, mxid, room_name,
                message, block, not bool(no_purge), force_purge)
        if room_del is None:
            click.echo("Room not deleted.")
            raise SystemExit(1)
        helper.output(room_del)
    else:
        click.echo("Abort.")


@room.command(name="purge-empty")
@optgroup.group(
    "Purge options", cls=MutuallyExclusiveOptionGroup,
    help="""Control how rooms are purged""")
@optgroup.option(
    "--no-purge", is_flag=True, default=False, show_default=True,
    help="""Prevent removing of all traces of the room from your
    database.""")
@optgroup.option(
    "--force-purge", is_flag=True, default=False, show_default=True,
    help="""Force a purge to go ahead even if there are local users still
    in the room. Do not use this unless a regular purge operation fails,
    as it could leave those users' clients in a confused state.""")
@click.option(
    "--v1", is_flag=True, default=False, show_default=True,
    help="""Use version 1 of the room delete API instead of version 2""")
@click.option(
    "--dry-run", is_flag=True, default=False,
    help="""Only show the rooms IDs that will be deleted""")
@click.option(
    "--batch-size", "--paginate", "-p", type=int, default=100,
    show_default=True,
    help="""How many rooms should be requested from the API one at a time.
    This option has no effect on how many empty rooms will be deleted.

    Increasing this is not necessary in most cases but useful if you have a
    lot of rooms on your homeserver.""")
@click.pass_obj
def purge_empty(helper, no_purge, force_purge, v1, dry_run, batch_size):
    """ Delete empty rooms (where 0 local members are currently in a room).
    """
    if no_purge and force_purge:
        click.echo("--force-purge will be ignored as --no-purge is set")

    empty_rooms_ids = []
    for room_list_response in helper.api.room_list_paginate(
            batch_size, None, "joined_local_members", True):
        found_empty_rooms = False

        if "rooms" not in room_list_response.keys():
            helper.log.warn("\"rooms\" key is missing from room list"
                            "response.")

        for room in room_list_response["rooms"]:
            room_id = room["room_id"]
            joined_local_members = room["joined_local_members"]
            if joined_local_members == 0:
                helper.log.debug(f"Added {room_id} to delete "
                                 f"(joined local members is "
                                 f"{joined_local_members})")
                empty_rooms_ids.append(room_id)
                found_empty_rooms = True
            else:
                helper.log.debug(f"Skipping {room_id} (joined local "
                                 f"members is {joined_local_members}, "
                                 f"not 0)")
                # very early cut off, hopefully always works and is never
                # wrong
                found_empty_rooms = False
                break

        # list is sorted by joined_local_members from smallest to biggest,
        # if there's no more where joined_local_members == 0 then just stop
        # early
        if not found_empty_rooms:
            helper.log.debug("No more empty rooms, stopping room list "
                             "fetching early.")
            break

    helper.output(empty_rooms_ids)
    if dry_run:
        click.echo("Empty room purge dry run. Rooms listed are considered "
                   "empty.", err=True)
        return

    sure = (
        helper.no_confirm or
        click.prompt("Are you sure you want to delete the listed empty "
                     "rooms? (y/N)", type=bool, default=False,
                     show_default=False)
    )
    if not sure:
        click.echo("Abort.", err=True)
        raise SystemExit(1)

    for room_id in empty_rooms_ids:
        if v1:
            result = helper.api.room_delete(
                    room_id, None, None, None, False, not no_purge,
                    force_purge)
            helper.output(result)
        else:
            result = helper.api.room_delete_v2(
                    room_id, None, None, None, False, not no_purge,
                    force_purge)
            helper.output(result)


@room.command(name="delete-status")
@optgroup.group(
        "Query type", cls=RequiredMutuallyExclusiveOptionGroup,
        help="Query room deletion status via either Room ID or Deletion ID"
)
@optgroup.option(
        "--room-id", "-r", type=str,
        help="""The Room ID to query the deletion status for""")
@optgroup.option(
        "--delete-id", "-d", type=str,
        help="""The Delete ID to query the deletion status for""")
@click.pass_obj
def delete_status(helper, room_id, delete_id):
    """ Get room deletion status via either the room ID or the delete ID.

    This requires the usage of the Room Delete v2 API. If you used v1 of the
    Room Delete API, this is irrelevant.
    """
    output = None
    if room_id:
        output = helper.api.room_delete_v2_status_by_room_id(
                room_id
        )
    if delete_id:
        output = helper.api.room_delete_v2_status_by_delete_id(
                delete_id
        )
    helper.output(output)


@room.command(name="search")
@click.argument("search-term", type=str)
@click.option(
    "--from", "-f", "from_", type=int, default=0, show_default=True,
    help="""Offset room listing by given number. This option is used for
    pagination.""")
@click.option(
    "--limit", "-l", type=int, default=100, show_default=True,
    help="Maximum amount of rooms to return.")
@click.option(
    "--sort", "-s", type=click.Choice(
        ["name", "canonical_alias", "joined_members", "joined_local_members",
         "version", "creator", "encryption", "federatable", "public",
         "join_rules", "guest_access", "history_visibility", "state_events"]),
    help="The method in which to sort the returned list of rooms.")
@click.option(
    "--reverse", "-r", is_flag=True, default=False,
    help="""Direction of room order. If set it will reverse the sort order of
    --order-by method.""")
@click.pass_context
def search_room_cmd(ctx, search_term, from_, limit, sort, reverse):
    """ An alias to `synadm room list -n <search-term>`.
    """
    ctx.invoke(list_room_cmd, from_=from_, limit=limit, name=search_term,
               sort=sort, reverse=reverse)


@room.command()
@click.argument("room_id", type=str)
@click.option(
    "--user-id", "-u", type=str,
    help="""By default the server admin (the caller) is granted power, but
    another user can optionally be specified.""")
@click.pass_obj
def make_admin(helper, room_id, user_id):
    """ Grant a user room admin permission.

    If the user is not in the room, and it is not publicly joinable, then
    invite the user.
    """

    mxid = helper.generate_mxid(user_id)
    out = helper.api.room_make_admin(room_id, mxid)
    helper.output(out)
    # note: this error appears to be synapse specific
    if "errcode" in out and (
            out["errcode"] == "M_UNKNOWN" and
            # main creator                additional creator case
            ("Creator" in out["error"] or "creator" in out["error"])
    ):
        helper.log.warn("In room version 12, room creators cannot be "
                        "assigned power levels. (MSC4289)")


@room.command()
@click.argument("room_id", type=str)
@click.option(
    "--block/--unblock", "-b/-u", type=bool, default=True, show_default=True,
    help="Specifies whether to block or unblock a room."
)
@click.pass_obj
def block(helper, room_id, block):
    """ Block or unblock a room.
    """
    out = helper.api.block_room(room_id, block)
    helper.output(out)


@room.command()
@click.argument("room_id", type=str)
@click.pass_obj
def block_status(helper, room_id):
    """ Get if a room is blocked, and who blocked it.
    """
    out = helper.api.room_block_status(room_id)
    helper.output(out)



# -*- coding: utf-8 -*-
# synadm
# Copyright (C) 2020-2025 Johannes Tiefenbacher
#
# synadm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# synadm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" CLI helpers and utilities"""

import os
import logging
import pprint
import json
import click
import yaml
import tabulate
from urllib.parse import urlparse
import dns.resolver
import re

from synadm import api


def humanize(data):
    """ Try to display data in a human-readable form:
    - Lists of dicts are displayed as tables.
    - Dicts are displayed as pivoted tables.
    - Lists are displayed as a simple list.
    """
    if isinstance(data, list) and len(data):
        if isinstance(data[0], dict):
            headers = {header: header for header in data[0]}
            return tabulate.tabulate(data, tablefmt="simple", headers=headers)
    if isinstance(data, list):
        return "\n".join(data)
    if isinstance(data, dict):
        return tabulate.tabulate(data.items(), tablefmt="plain")
    return str(data)


class APIHelper:
    """ API client enriched with CLI-level functions, used as a proxy to the
    client object.
    """

    FORMATTERS = {
        "pprint": pprint.pformat,
        "json": lambda data: json.dumps(data, indent=4),
        "minified": lambda data: json.dumps(data, separators=(",", ":")),
        "yaml": yaml.dump,
        "human": humanize
    }

    # consider this kind of like defaults
    CONFIG = {
        "user": "",
        "token": "",
        "protocol": "http",
        "base_url": "http://localhost:8008",
        "admin_path": "/_synapse/admin",
        "matrix_path": "/_matrix",
        "timeout": 30,
        "server_discovery": "well-known",
        "homeserver": "auto-retrieval",
        "ssl_verify": True
    }

    def __init__(self, config_path, verbose, no_confirm, output_format_cli):
        self.config = APIHelper.CONFIG.copy()
        self.config_path = os.path.expanduser(config_path)
        self.no_confirm = no_confirm
        self.api = None
        self.init_logger(verbose)
        self.requests_debug = False
        if verbose >= 3:
            self.requests_debug = True
        self.output_format_cli = output_format_cli  # override from cli

    def init_logger(self, verbose):
        """ Log both to console (defaults to WARNING) and file (DEBUG).
        """
        log_path = os.path.expanduser("~/.local/share/synadm/debug.log")
        os.makedirs(os.path.dirname(log_path), exist_ok=True)
        log = logging.getLogger("synadm")
        log.setLevel(logging.DEBUG)
        file_handler = logging.FileHandler(log_path, encoding="utf-8")
        file_handler.setLevel(logging.DEBUG)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(
            logging.DEBUG if verbose > 1 else
            logging.INFO if verbose == 1 else
            logging.WARNING
        )
        file_formatter = logging.Formatter(
            "%(asctime)s %(name)-8s %(levelname)-7s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S"
        )
        console_formatter = logging.Formatter("%(levelname)-5s %(message)s")
        console_handler.setFormatter(console_formatter)
        file_handler.setFormatter(file_formatter)
        log.addHandler(console_handler)
        log.addHandler(file_handler)
        self.log = log

    def _set_formatter(self, _output_format):
        for name, formatter in APIHelper.FORMATTERS.items():
            if name.startswith(_output_format):
                self.output_format = name
                self.formatter = formatter
                break
        self.log.debug("Formatter in use: %s - %s", self.output_format,
                       self.formatter)
        return True

    def load(self):
        """ Load the configuration and initialize the client.

        Returns: A boolean telling if the config is complete enough and
        loaded.
        """
        try:
            with open(self.config_path) as handle:
                self.config.update(yaml.load(handle, Loader=yaml.SafeLoader))
        except Exception as error:
            self.log.error("%s while reading configuration file", error)
        for key, value in self.config.items():

            if key == "ssl_verify" and not isinstance(value, bool):
                self.log.error("Config value error: %s, %s must be boolean",
                               key, value)

            if not value and not isinstance(value, bool):
                self.log.error("Config entry missing: %s, %s", key, value)
                return False
            else:
                if key == "token":
                    self.log.debug("Config entry read. %s: REDACTED", key)
                else:
                    self.log.debug("Config entry read. %s: %s", key, value)
        if self.output_format_cli:  # we have a cli output format override
            self._set_formatter(self.output_format_cli)
        else:  # we use the configured default output format
            self._set_formatter(self.config["format"])
        self.api = api.SynapseAdmin(
            self.log,
            self.config["user"], self.config["token"],
            self.config["protocol"], self.config["base_url"],
            self.config["admin_path"],
            self.config["timeout"], self.requests_debug,
            self.config["ssl_verify"]
        )
        self.matrix_api = api.Matrix(
            self.log,
            self.config["user"], self.config["token"],
            self.config["protocol"], self.config["base_url"],
            self.config["matrix_path"],
            self.config["timeout"], self.requests_debug,
            self.config["ssl_verify"]
        )
        self.misc_request = api.MiscRequest(
            self.log,
            self.config["timeout"], self.requests_debug,
            self.config["ssl_verify"]
        )
        return True

    def write_config(self, config):
        """ Write a new version of the configuration to file.
        """
        try:
            os.makedirs(os.path.dirname(self.config_path), exist_ok=True)
            with open(self.config_path, "w") as handle:
                yaml.dump(config, handle, default_flow_style=False,
                          allow_unicode=True)
            if os.name == "posix":
                click.echo("Restricting access to config file to user only.")
                os.chmod(self.config_path, 0o600)
            else:
                click.echo(f"Unsupported OS, please adjust permissions of "
                           f"{self.config_path} manually")

            return True
        except Exception as error:
            self.log.error("%s trying to write configuration", error)
            return False

    def output(self, data):
        """ Output data object using the configured formatter.
        """
        click.echo(self.formatter(data))

    def retrieve_homeserver_name(self, protocol=None, uri=None):
        """Try to retrieve the homeserver name.

        When homeserver is set in the config already, it's just returned and
        nothing is tried to be fetched automatically. If not, either the
        location of the Federation API is looked up via a .well-known resource
        or a DNS SRV lookup. This depends on the server_discovery setting in
        the config. Finally the Federation API is used to retrieve the
        homeserver name.

        Args:
            protocol (string): Either "http" or "unix". Defaults to
                configuration.
            uri (string): proto://name:port or proto://fqdn:port, or an
                absolute path without any protocol specifications when the
                protocol is set to "unix". Defaults to configuration's
                base_url setting.

        Returns:
            string: hostname, FQDN or DOMAIN; or None on errors.
        """
        # this exists for the purpose of local media deletion (and maybe a
        # couple of other things), which for some reason requires the HS
        # name
        #
        # the way that's found is /_matrix/key/v2/server (server-server
        # API), with a messy domain delegation resolving approach
        uri = uri if uri else self.config["base_url"]
        protocol = protocol if protocol else self.config["protocol"]
        echo = self.log.info if self.no_confirm else click.echo
        if self.config["homeserver"] != "auto-retrieval":
            return self.config["homeserver"]

        # directly accessible synapse method (so key to name)
        if ("localhost" in self.config["base_url"] or
                "127.0.0.1" in self.config["base_url"] or
                "[::1]" in self.config["base_url"] or
                self.config["protocol"] == "unix"):
            echo(
                "Trying to fetch homeserver name via localhost..."
            )
            return self.matrix_api.server_name_keys_api(
                self.config["protocol"], self.config["base_url"]
            )
        # well-known delegation, then key to name
        elif self.config["server_discovery"] == "well-known":
            echo(
                "Trying to fetch federation URI via well-known resource..."
            )
            federation_uri = self.misc_request.federation_uri_well_known(
                protocol, uri
            )
            if not federation_uri:
                return None
            # hard coding "http" as protocol because "unix" method is dealt
            # above
            return self.matrix_api.server_name_keys_api("http", federation_uri)
        # dns delegation (deprecated), then key to name
        elif self.config["server_discovery"] == "dns":
            echo(
                "Trying to fetch federation URI via DNS SRV record..."
            )
            hostname = urlparse(uri).hostname
            try:
                record = dns.resolver.query(
                    "_matrix._tcp.{}".format(hostname),
                    "SRV"
                )
            except Exception as error:
                self.log.error(
                    "resolving Matrix delegation for %s: %s: %s",
                    hostname, type(error).__name__, error
                )
            else:
                federation_uri = "https://{}:{}".format(
                    record[0].target, record[0].port
                )
                # hard coding "http" as protocol because "unix" method is
                # dealt above
                return self.matrix_api.server_name_keys_api("http",
                                                            federation_uri)
        else:
            self.log.error("Unknown server_discovery mode. "
                           "Launch synadm config!")
        # edge case potential: no domain delegation, and doesn't match the
        # first condition in this if-tree
        return None

    def generate_mxid(self, user_id):
        """ Checks whether the given user ID is an MXID already or else
        generates it from the passed string and the homeserver name fetched
        via the retrieve_homeserver_name method.

        Args:
            user_id (string): User ID given by user as command argument.

        Returns:
            string: the fully qualified Matrix User ID (MXID) or None if the
                user_id parameter is None or no regex matched.
        """
        if user_id is None:
            self.log.debug("Missing input in generate_mxid().")
            return None
        elif re.match(r"^@[-./=+\w]+:[-\[\].:\w]+$", user_id):
            self.log.debug("A proper MXID was passed.")
            return user_id
        elif re.match(r"^@?[-./=+\w]+:?$", user_id):
            self.log.debug("A proper localpart was passed, generating MXID "
                           "for local homeserver.")
            localpart = re.sub("[@:]", "", user_id)
            mxid = "@{}:{}".format(localpart, self.retrieve_homeserver_name())
            return mxid
        else:
            self.log.error("Neither an MXID nor a proper localpart was "
                           "passed.")
            return None



# -*- coding: utf-8 -*-
# synadm
# Copyright (C) 2023-2025 Johannes Tiefenbacher
#
# synadm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# synadm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module holds the `raw` command only."""

import click

from synadm import cli
from synadm.cli._common import common_opts_raw_command, data_opts_raw_command


@cli.root.command(name="raw")
@common_opts_raw_command
@data_opts_raw_command
@click.pass_obj
def raw_request_cmd(helper, endpoint, method, data, data_file):
    """ Issue a custom request to the Synapse Admin API.

    The endpoint argument is the part of the URL _after_ the configured
    "Synapse base URL" and "Synapse Admin API path" (see ``synadm config``).
    A get request to the "Query User Account API" would look like this:
    ``synadm raw v2/users/%40testuser%3Aexample.org``. URL encoding must be
    handled at this point. Consider enabling debug outputs via synadm's global
    flag ``-vv``
    """
    if data_file:
        data = data_file.read()

    raw_request = helper.api.raw_request(endpoint, method, data)

    if helper.no_confirm:
        if raw_request is None:
            raise SystemExit(1)
        helper.output(raw_request)
    else:
        if raw_request is None:
            click.echo("The Admin API's response was empty or JSON data "
                       "could not be loaded.")
            raise SystemExit(1)
        else:
            helper.output(raw_request)
